/*
  XPLDirect.h - Library for serial interface to Xplane SDK.
  Created by Michael Gerlicher,  September 2020.

  To report problems, download updates and examples, suggest enhancements or get technical support, please visit my patreon page:
     
     www.patreon.com/curiosityworkshop
	
 
*/

#ifndef XPLDirect_h
#define XPLDirect_h


#define XPLDIRECT_BAUDRATE 115200
#define XPLDIRECT_PACKETHEADER  '<'
#define XPLDIRECT_PACKETTRAILER '>'
#define XPLDIRECT_MAXDATAREFS_PC 100
#define XPLDIRECT_MAXCOMMANDS_PC 100
#define XPLDIRECT_MAXDATAREFS_ARDUINO 20
#define XPLDIRECT_MAXCOMMANDS_ARDUINO 20
#define XPLDEVICES_MAXDEVICES 20

#define XPLMAX_PACKETSIZE 150
#define XPL_TIMEOUT_SECONDS 3

#define XPLERROR                   'E'   // %s         general error
#define XPLRESPONSE_NAME           '0'       
#define XPLRESPONSE_DATAREF        '3'   // %3.3i%s    dataref handle, dataref name 
#define XPLRESPONSE_COMMAND        '4'   // %3.3i%s    command handle, command name
#define XPLCMD_PRINTDEBUG          '1'
#define XPLCMD_RESET               '2'
#define XPLCMD_SENDNAME            'a'
#define XPLREQUEST_REGISTERDATAREF 'b'   //  %1.1i%2.2i%5.5i%s    RWMode, array index (0 for non array datarefs), divider to decrease resolution, dataref name
#define XPLREQUEST_REGISTERCOMMAND 'm'  // just the name of the command to register
#define XPLREQUEST_NOREQUESTS      'c'   // nothing to request
#define XPLREQUEST_REFRESH         'd'	//  the plugin will call this once xplane is loaded in order to get fresh updates from arduino handles that write

#define XPLCMD_DUMPREGISTRATIONS   'Z'  // for debug purposes only
#define XPLCMD_DATAREFUPDATE       'e'
#define XPLCMD_SENDREQUEST         'f'
#define XPLCMD_DEVICEREADY         'g'
#define XPLCMD_DEVICENOTREADY      'h'
#define XPLCMD_COMMANDSTART         'i'
#define XPLCMD_COMMANDEND           'j'

#define XPL_READ		1
#define XPL_WRITE       2
#define XPL_READWRITE	3

#define XPL_DATATYPE_INT    1
#define XPL_DATATYPE_FLOAT  2
#define XPL_DATATYPE_STRING 3

#define XPL_RETURN_TIME_MILLIS 20
#define XPL_RETURN_TIME_MILLIS_UNREGISTERED 1
#define XPL_COMMAND_MAX_UPDATE_RATE_MILLIS  300         // good for debounce


class XPLDirect
{
  public:

    XPLDirect(Stream*); 
    void begin(char *devicename);     // parameter is name of your device for reference 
  
    int connectionStatus(void);
    
    int registerDataRef(const char *, int, unsigned int, float, long int *);           // single int dataref.  name, Mode, rate, divider, Value
    int registerDataRef(const char*, int, unsigned int, float, long int*, int);        // array int dataref
    int registerDataRef(const char*, int, unsigned int, float, float*);                // single float dataref
    int registerDataRef(const char*, int, unsigned int, float, float*, int);           // array float dataref
    int registerCommand(const char* commandName, int* value);                           
    int sendReadyCommand(void);
    int getCurrentData(int, int*);
    int sendDebugMessage(char*);
    int allDataRefsRegistered(void);
    void dumpRegistrations(void);						// a debug function
   
    int xloop(void);
  
  private:
      void _processSerial();
      void _processPacket();

      void _sendPacket(int command, int handle, int value);         // for ints
      void _sendPacket(int command, int handle, float value);       // for floats
      void _sendPacket(int command, int handle);                    // just a command with a handle
      void _sendPacket(int command, char* str);                     // send a string
   
      void _transmitPacket();

      void _sendname();
      int _getResponse(int *);

      int _getHandleFromFrame();
      int _getPayloadFromFrame(long int *);
      int _getPayloadFromFrame(float*);

      Stream* streamPtr;

    
      char _receiveBuffer[XPLMAX_PACKETSIZE]; int _receiveBufferPosition;
      char _sendBuffer[XPLMAX_PACKETSIZE];
  
     

      int _pin;
      int _connectionStatus;

    char * _deviceName;

    struct _dataRefStructure
    {
        int dataRefHandle;
        int dataRefRWType;               //XPL_READ, XPL_WRITE, XPL_READWRITE
        int dataRefVARType;              // XPL_DATATYPE_INT 1, XPL_DATATYPE_FLOAT  2   XPL_DATATYPE_STRING 3
        float divider;                     // tell the host to reduce resolution by dividing then remultiplying by this number to reduce traffic.   (ie 1, 10, 100, 1000)
        int forceUpdate;                    // in case xplane plugin asks for a refresh
        unsigned long updateRate;        // maximum update rate in milliseconds, 0 = every change
        unsigned long lastUpdateTime;
        const char* dataRefName;
        void* latestValue;
        long int lastSentIntValue;
        float lastSentFloatValue;
        int arrayIndex;                 // for datarefs that speak in arrays

    };

    struct _commandStructure
    {
        int commandHandle;
        const char* commandName;
        long int lastUpdateTime;
        int *latestValue;
        int lastSentValue;

    };

    _dataRefStructure *_dataRefs[XPLDIRECT_MAXDATAREFS_ARDUINO];
    _commandStructure *_commands[XPLDIRECT_MAXCOMMANDS_ARDUINO];

    int _dataRefsCount;
    int _commandsCount;
    int _allDataRefsRegistered;
    
   
};

#endif
